
/////////////// heightfield parameter ////////////////////////


// view projection matrix
float4x4 xViewProjection;


// translation in world coordinates that is applied to the whole heightfield
float3 xHeightfieldTranslation;


// factor with which the heightmap height values are multiplied, in world coordinates
float xMaxAlteringHeight;


// factor multiplied to the index (x,y number) of each instanced patch, to get the texture 
// coords on the heightmap 
float2 xHeightfieldPixelOffset;


// world size of a patch (patch is quadratic)
float xPatchWidth;


// GLOBAL TEXTURES: each pixel defines a value for the grid
//=============================================================================================

// heightfield texture: used to perturbate the height of the grid
uniform Texture xHeightfield;
sampler HeightTextureSampler = sampler_state
{
	texture = <xHeightfield>;
	//magfilter = LINEAR;
	//minfilter = LINEAR;
	//mipfilter = LINEAR;
	magfilter = NONE;
	minfilter = NONE;
	mipfilter = NONE;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};



/////////////// shadowmapping parameter //////////////////////


uniform float4x4	xView;
uniform float4x4	xProjection;
uniform float4x4	xWorld;
uniform float4x4	xMeshBoneMatrix;
uniform float		xMinDistanceLight;
uniform float		xMaxDepth;
uniform float3		xLightDirection;



uniform int			xNumShadowMaps;
uniform float2		xNearFarRange;
uniform float2		xCamNearFar;

uniform int			xMapResolution;

uniform int			xLogBaseSplitNum;


uniform float4x4	xLightViewProjection0;
uniform float4x4	xLightViewProjection1;
uniform float4x4	xLightViewProjection2;
uniform float4x4	xLightViewProjection3;
uniform float4x4	xLightViewProjection4;
uniform float4x4	xLightViewProjection5;
uniform float4x4	xLightViewProjection6;
uniform float4x4	xLightViewProjection7;

uniform Texture xShadowMap0;
sampler ShadowMapSampler0 = sampler_state
{
	texture = <xShadowMap0>;
	//magfilter = POINT;
	//minfilter = POINT;
	//mipfilter = POINT;
	magfilter = Anisotropic;
	minfilter = Anisotropic;
	maxanisotropy = 16;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};


uniform Texture xShadowMap1;
sampler ShadowMapSampler1 = sampler_state
{
	texture = <xShadowMap1>;
	magfilter = POINT;
	minfilter = POINT;
	mipfilter = POINT;
	//magfilter = Anisotropic;
	//minfilter = Anisotropic;
	//maxanisotropy = 16;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};


uniform Texture xShadowMap2;
sampler ShadowMapSampler2 = sampler_state
{
	texture = <xShadowMap2>;
	magfilter = POINT;
	minfilter = POINT;
	mipfilter = POINT;
	//magfilter = Anisotropic;
	//minfilter = Anisotropic;
	//maxanisotropy = 16;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};

uniform Texture xShadowMap3;
sampler ShadowMapSampler3 = sampler_state
{
	texture = <xShadowMap3>;
	magfilter = POINT;
	minfilter = POINT;
	mipfilter = POINT;
	//magfilter = Anisotropic;
	//minfilter = Anisotropic;
	//maxanisotropy = 16;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};


uniform Texture xShadowMap4;
sampler ShadowMapSampler4 = sampler_state
{
	texture = <xShadowMap4>;
	magfilter = POINT;
	minfilter = POINT;
	mipfilter = POINT;
	//magfilter = Anisotropic;
	//minfilter = Anisotropic;
	//maxanisotropy = 16;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};

/*
uniform Texture xShadowMap5;
sampler ShadowMapSampler5 = sampler_state
{
	texture = <xShadowMap5>;
	magfilter = POINT;
	minfilter = POINT;
	mipfilter = POINT;
	//magfilter = Anisotropic;
	//minfilter = Anisotropic;
	//maxanisotropy = 16;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};


uniform Texture xShadowMap6;
sampler ShadowMapSampler6 = sampler_state
{
	texture = <xShadowMap6>;
	magfilter = POINT;
	minfilter = POINT;
	mipfilter = POINT;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};


uniform Texture xShadowMap7;
sampler ShadowMapSampler7 = sampler_state
{
	texture = <xShadowMap7>;
	magfilter = POINT;
	minfilter = POINT;
	mipfilter = POINT;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};
*/



// input structs

// normal Vertex Shader Input
struct VS_input {
	float4 Position		: POSITION;
	float2 TexCoords	: TEXCOORD0;
};


struct SScene_v2p
{
    float4 Position             : POSITION;
    float4 Position3DViewspace  : TEXCOORD0;
    float4 Position3D           : TEXCOORD1;
    float3 Normal				: TEXCOORD2;
    float4 DynamicLightFactors	: TEXCOORD3;
};



struct SScene_p2f
{
	float4 Color		: COLOR0;
};







// shadowed heightfield vertex shader
/*
Fast_Z_PS_input Fast_Z_Pass_VS( VS_input input, float2 PatchIndex : TEXCOORD2 )
{
	// 0) declare Output and initialize all variables to 0
	SScene_v2p Output = (SScene_v2p)0;
	
	// 1) get heightfield texture value & compute height value
	float2 globalTexCoords = input.TexCoords + PatchIndex * xHeightfieldPixelOffset;
	float texColor = tex2Dlod(HeightTextureSampler, float4(globalTexCoords,0.0f,1.0f)).x;
	float3 h = {0.0f, xMaxAlteringHeight * texColor, 0.0f};
	
	// 3) compute translation
	float3 t = {PatchIndex.x * xPatchWidth, 0.0f, PatchIndex.y * xPatchWidth};

	
	// 4) store output
	float4 finalPos4 = float4(input.Position + h + t + xHeightfieldTranslation, 1.0f);
	Output.Position = mul(finalPos4, xViewProjection);
	return Output;
}
*/




SScene_v2p shadowedHeightfieldVertexShader( VS_input input, float2 PatchIndex : TEXCOORD2 )
{
    
    // 1) get heightfield texture value & compute height value
	float2 globalTexCoords = input.TexCoords + PatchIndex * xHeightfieldPixelOffset;
	float texColor = tex2Dlod(HeightTextureSampler, float4(globalTexCoords,0.0f,1.0f)).x;
	float3 h = {0.0f, xMaxAlteringHeight * texColor, 0.0f};
	
	// 3) compute translation
	float3 t = {PatchIndex.x * xPatchWidth, 0.0f, PatchIndex.y * xPatchWidth};

	
	// 4) store output
	float4 finalPos4 = float4(input.Position + h + t + xHeightfieldTranslation, 1.0f);

   
    
    SScene_v2p Output = (SScene_v2p)0;
    
    // compute projected and 3D positions
    Output.Position3D = finalPos4;
    Output.Position3DViewspace = mul(Output.Position3D, xView);
    Output.Position = mul(Output.Position3DViewspace, xProjection);
    
    // offset z-value to overlay textured scene with shadowed scene in backbuffer
    Output.Position.z -= 0.0014;
    
    // pass on normal to pixel shader
	//Output.Normal = normalize(mul(inNormal, xWorld));
	
	
    return Output;    
}






float4 computeShadowMap(sampler shadowMapSampler, 
                        Matrix shadowMapProjection, 
                        float4 Position3D, 
                        float errorCorrectionParameter) {

	float4 ShadowMapSamplingPos = mul(Position3D, shadowMapProjection);   
    
    // clamp position from [-1, 1] to [0, 1]	
	ShadowMapSamplingPos.xy = ShadowMapSamplingPos.xy * float2(0.5, -0.5) + float2(0.5, 0.5);
	
	// define default color
    float4 color = float4(1.0, 0.0, 0.0, 0.0);
   
	// lies current pixel within shadow map texture?
	//float cosNL = dot(Normal, xLightDirection);
	//float sinNL = Cross(Normal, -xLightDirection);
	//if (cosNL >= -0.0) {
	//	color = float4( 0.0, 0.0, 0.0, 0.0);
	//	color.w = 0.5;
	//} else 
	
	if ((saturate(ShadowMapSamplingPos.x) == ShadowMapSamplingPos.x) && (saturate(ShadowMapSamplingPos.y) == ShadowMapSamplingPos.y))
	{		
		// compute real distance of position
		float RealDistance = (ShadowMapSamplingPos.z - xMinDistanceLight) / xMaxDepth;

		// define delta tex coord of one pixel step
		float delta = 1.0 / float(xMapResolution);
			
		// define new default color	(rgb=shadow color, alpha will be set later);
		color = float4( 0.0, 0.0, 0.0, 0.0);
		
		
		
		
		//if (RealDistance + errorCorrectionParameter > tex2D(shadowMapSampler, ShadowMapSamplingPos.xy).x) {	
			
			int kernelWidth = 1;
		
			float inShadow = 0;
		
			for (int x = -kernelWidth; x <= kernelWidth; x+=2) {
				for (int y = -kernelWidth; y <= kernelWidth; y+=2) {
					float2 pDelta = float2(x * delta, y * delta);			
					
					if (RealDistance + errorCorrectionParameter > tex2D(shadowMapSampler, (ShadowMapSamplingPos.xy + pDelta)).x) {
						inShadow += 1;
					}
				}
			}
			
			//float n = (kernelWidth * 2 + 1) * (kernelWidth * 2 + 1);
			float n = (kernelWidth + 1) * (kernelWidth + 1);
			//float n = (kernelWidth * 0.5f + 1) * (kernelWidth * 0.5f + 1);
			color.w = 0.4 * (inShadow /n);
			
		//}
		
	}
	return color;
}





SScene_p2f shadowedHeightfieldPixelShader(SScene_v2p PSIn)
{
	SScene_p2f Output = (SScene_p2f)0;

	// define default color
	float4 color = float4( 0.0, 0.0, 1.0, 0.5 );


	float pixelDist3DViewspace = abs(PSIn.Position3DViewspace.z / PSIn.Position3DViewspace.w) - xCamNearFar.x;

	float totalSplitVolumesLenght = xNearFarRange.y - xNearFarRange.x; 
	

	if (pixelDist3DViewspace < totalSplitVolumesLenght) {

		

		// find out which map index should be used 
		///////////////////////////////////////////
		

		
		// LOGARITHMIC SPLITTING:
		float deltaSplitLen = totalSplitVolumesLenght / xLogBaseSplitNum;
		int mapIndex = floor(pixelDist3DViewspace / deltaSplitLen);


		if (mapIndex < 1) {
			color = computeShadowMap(ShadowMapSampler0, xLightViewProjection0, PSIn.Position3D,   -0.000000); // tuned urban -0.00001
			//color.x = 1.0f;
		} else
		if (mapIndex < 3) {
			color = computeShadowMap(ShadowMapSampler1, xLightViewProjection1, PSIn.Position3D,   -0.000000);  // tuned urban
			//color.y = 1.0f;
		} else
		if (mapIndex < 6) { 
			color = computeShadowMap(ShadowMapSampler2, xLightViewProjection2, PSIn.Position3D,   -0.000000);  // 
			//color.z = 1.0f;
		} else
		if (mapIndex < 10) {
			color = computeShadowMap(ShadowMapSampler3, xLightViewProjection3, PSIn.Position3D,   -0.000000);
			//color.x = 1.0f;
			//color.y = 1.0f;
		} else
		if (mapIndex <=15) {
			color = computeShadowMap(ShadowMapSampler4, xLightViewProjection4, PSIn.Position3D, -0.000012);
			//color.x = 1.0f;
			//color.z = 1.0f;
		}
		
	} else {
		color = (1, 1, 1, 1);
	}
	Output.Color = color;
    return Output;
}





// technique for rendering of the shadow map
technique shadowedHeightfield
{
    pass
    {
        VertexShader = compile vs_3_0 shadowedHeightfieldVertexShader();
        PixelShader = compile ps_3_0 shadowedHeightfieldPixelShader();
    }
}
