// ========================================================================================== //
//             ___  _   _    __    __  __    __    ____  __    ____  _____  _  _              //
//            / __)( )_( )  /__\  (  \/  )  /__\  ( ___)(  )  ( ___)(  _  )( \( )             //
//           ( (__  ) _ (  /(__)\  )    (  /(__)\  )__)  )(__  )__)  )(_)(  )  (              //
//            \___)(_) (_)(__)(__)(_/\/\_)(__)(__)(____)(____)(____)(_____)(_)\_)             //
//            -------------------------------------------- G A M E   E N G I N E              //
//                                                                                            //
// ========================================================================================== //
// Developed and implemented by:                                                              //
//     Oskam Thomas                                                                           //
//     Mader Theodor                                                                          //
//                                                                                            //
// Copyright 2007-2009. All rights reserved.                                                  //
// ========================================================================================== //


using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using System.Text;
using HFShader.Tools;

namespace HFShader.Rendering
{


    // config parameter names
    public static class HFPNames
    {
        public const string AlphaStartEnd = "AlphaStartEnd";
        public const string WorldHeight = "WorldHeightFactor";
        public const string WorldWidth = "WorldWidthFactor";

        public const string Translation = "Translation";
        public const string LightDirection = "LightDirection";

        public const string RedChnTex = "RedChannelTexture";
        public const string GreenChnTex = "GreenChannelTexture";
        public const string BlueChnTex = "BlueChannelTexture";
        public const string AlphaChnTex = "AlphaChannelTexture";

        public const string HeightTex = "HeightmapTexture";
        public const string ChannelTex = "ColorChannelTexture";
        public const string NormalTex = "NormalTexture";

        public const string RedStretch = "RedStretchFactor";
        public const string GreenStretch = "GreenStretchFactor";
        public const string BlueStretch = "BlueStretchFactor";
        public const string AlphaStretch = "AlphaStretchFactor";

        public const string RenderMode = "RenderMode";
    }


    public class HeightfieldLoader
    {


        // separator character
        private string mSeparator = " ";



        // constructor
        public HeightfieldLoader() { 
        }




        /// <summary>
        /// Stores the heightfield parameter in hfp to a config file.
        /// </summary>
        /// <param name="hfp">Heightfield Parameter.</param>
        /// <param name="filePathWrtContent">path where to store (path starts automatically at "Content/"</param>
        public void storeHeightfieldSetup(HeightfieldParameter hfp, string filePathWrtContent)
        {

            // Delete the file if it exists.
            if (System.IO.File.Exists("Content/" + filePathWrtContent + ".hfp"))
            {
                System.IO.File.Delete("Content/" + filePathWrtContent + ".hfp");
            }


            // buffer
            byte[] info;

            
            // Create the file.
            using (System.IO.FileStream fs = System.IO.File.Create("Content/" + filePathWrtContent + ".hfp"))
            {

                // write header
                //////////////////

                info = new System.Text.UTF8Encoding(true).GetBytes("Heightfield Loader File Version 1.0\n");
                fs.Write(info, 0, info.Length);
                
                info = new System.Text.UTF8Encoding(true).GetBytes("   Chamaeleon Engine Heightfield Description File\n");
                fs.Write(info, 0, info.Length);
                
                info = new System.Text.UTF8Encoding(true).GetBytes("\n");
                fs.Write(info, 0, info.Length);



                // write information
                //////////////////////

                // Alpha start end
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.AlphaStartEnd, hfp.Alpha_Start_End));
                fs.Write(info, 0, info.Length);

                // world height 
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.WorldHeight, hfp.HeightfieldWorldHeight));
                fs.Write(info, 0, info.Length);

                // world width 
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.WorldWidth, hfp.HeightfieldWorldWidth));
                fs.Write(info, 0, info.Length);




                // translation
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.Translation, hfp.Translation));
                fs.Write(info, 0, info.Length);


                // light direction
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.LightDirection, hfp.LightDirection));
                fs.Write(info, 0, info.Length);




                // red channel texture 
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.RedChnTex, hfp.RedChannelTexture));
                fs.Write(info, 0, info.Length);
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.RedStretch, hfp.RedStretchFactor));
                fs.Write(info, 0, info.Length);

                // green channel texture 
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.GreenChnTex, hfp.GreenChannelTexture));
                fs.Write(info, 0, info.Length);
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.GreenStretch, hfp.GreenStretchFactor));
                fs.Write(info, 0, info.Length);
                
                // blue channel texture 
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.BlueChnTex, hfp.BlueChannelTexture));
                fs.Write(info, 0, info.Length);
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.BlueStretch, hfp.BlueStretchFactor));
                fs.Write(info, 0, info.Length);

                // alpha channel texture 
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.AlphaChnTex, hfp.AlphaChannelTexture));
                fs.Write(info, 0, info.Length);
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.AlphaStretch, hfp.AlphaStretchFactor));
                fs.Write(info, 0, info.Length);



                // heightmap texture 
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLineNStore(HFPNames.HeightTex, filePathWrtContent + "_" + HFPNames.HeightTex, hfp.HeightfieldTexture, ImageFileFormat.Dds));
                fs.Write(info, 0, info.Length);


                // colorchannel texture
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLineNStore(HFPNames.ChannelTex, filePathWrtContent + "_" + HFPNames.ChannelTex, hfp.ColorChannelTexture, ImageFileFormat.Tga));
                fs.Write(info, 0, info.Length);

                // normal texture
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLineNStore(HFPNames.NormalTex, filePathWrtContent + "_" + HFPNames.NormalTex, hfp.NormalTexture, ImageFileFormat.Tga));
                fs.Write(info, 0, info.Length);





                // render mode
                info = new System.Text.UTF8Encoding(true).GetBytes(getParamLine(HFPNames.RenderMode, (int)hfp.RenderMode));
                fs.Write(info, 0, info.Length);
            }
        }






        public HeightfieldParameter loadHeightfieldSetup(string filePathWrtContent, ContentManager contentManager, GraphicsDevice graphicsDevice)
        {
            HeightfieldParameter hfp = null;

            // check file
            if (!System.IO.File.Exists("Content/" + filePathWrtContent + ".hfp")) return hfp;


            // prepare variables
            string s = null;
            

            // open file
            using (System.IO.StreamReader sr = System.IO.File.OpenText("Content/" + filePathWrtContent + ".hfp"))
            {
                hfp = new HeightfieldParameter();

                // read header
                s = sr.ReadLine();
                s = sr.ReadLine();
                s = sr.ReadLine();


                // read parameter
                ///////////////////////

                // alpha start end
                s = sr.ReadLine();
                hfp.Alpha_Start_End = parseParameterV2(HFPNames.AlphaStartEnd, s);
                
                // world heithg
                s = sr.ReadLine();
                hfp.HeightfieldWorldHeight = parseParameterF(HFPNames.WorldHeight, s);

                // world width
                s = sr.ReadLine();
                hfp.HeightfieldWorldWidth = parseParameterF(HFPNames.WorldWidth, s);




                // translation
                s = sr.ReadLine();
                hfp.Translation = parseParameterV3(HFPNames.Translation, s);

                // light direction
                s = sr.ReadLine();
                hfp.LightDirection = parseParameterV3(HFPNames.LightDirection, s);




                // color channel textures
                s = sr.ReadLine();
                hfp.RedChannelTexture = parseParameterS(HFPNames.RedChnTex, s); 
                s = sr.ReadLine();
                hfp.RedStretchFactor = parseParameterF(HFPNames.RedStretch, s);

                s = sr.ReadLine();
                hfp.GreenChannelTexture = parseParameterS(HFPNames.GreenChnTex, s);
                s = sr.ReadLine();
                hfp.GreenStretchFactor = parseParameterF(HFPNames.GreenStretch, s);

                s = sr.ReadLine();
                hfp.BlueChannelTexture = parseParameterS(HFPNames.BlueChnTex, s);
                s = sr.ReadLine();
                hfp.BlueStretchFactor = parseParameterF(HFPNames.BlueStretch, s);

                s = sr.ReadLine();
                hfp.AlphaChannelTexture = parseParameterS(HFPNames.AlphaChnTex, s);
                s = sr.ReadLine();
                hfp.AlphaStretchFactor = parseParameterF(HFPNames.AlphaStretch, s);



                // heightmap texture
                s = sr.ReadLine();
                hfp.HeightfieldTexture = parseParameterT(HFPNames.HeightTex, s, contentManager, graphicsDevice);

                // color channel texture
                s = sr.ReadLine();
                hfp.ColorChannelTexture = parseParameterT(HFPNames.ChannelTex, s, contentManager, graphicsDevice);

                // normal texture
                s = sr.ReadLine();
                hfp.NormalTexture = parseParameterT(HFPNames.NormalTex, s, contentManager, graphicsDevice);



                // render mode
                s = sr.ReadLine();
                hfp.RenderMode = parseParameterI(HFPNames.RenderMode, s);
            }

            return hfp;
        }




        public void exportAsObj(List<ITexTriangle> triangles, string filePathWrtContent)
        {
            // Delete the file if it exists.
            if (System.IO.File.Exists("Content/" + filePathWrtContent + ".obj"))
            {
                System.IO.File.Delete("Content/" + filePathWrtContent + ".obj");
            }


            // buffer
            byte[] info;


            // Create the file.
            using (System.IO.FileStream fs = System.IO.File.Create("Content/" + filePathWrtContent + ".obj"))
            {

                // write header
                //////////////////

                info = new System.Text.UTF8Encoding(true).GetBytes("# " + "Heightfield OBJ Export v1.0\n");
                fs.Write(info, 0, info.Length);

                info = new System.Text.UTF8Encoding(true).GetBytes("# " + "   Created with Chamaeleon Engine. No guarantees whatsoever.\n");
                fs.Write(info, 0, info.Length);

                info = new System.Text.UTF8Encoding(true).GetBytes("\n");
                fs.Write(info, 0, info.Length);

                

                // group line
                info = new System.Text.UTF8Encoding(true).GetBytes("g default\n");
                fs.Write(info, 0, info.Length);

                //// write information
                ////////////////////////

                // write vertices
                //info = new System.Text.UTF8Encoding(true).GetBytes("# " + (triangles.Count * 3).ToString() + " Vertices\n");
                //fs.Write(info, 0, info.Length);
                foreach (Tools.ITexTriangle t in triangles)
                {
                    info = new System.Text.UTF8Encoding(true).GetBytes("v " + t.P0.X.ToString() + " " + t.P0.Y.ToString() + " " + t.P0.Z.ToString() + "\n");
                    fs.Write(info, 0, info.Length);
                    info = new System.Text.UTF8Encoding(true).GetBytes("v " + t.P1.X.ToString() + " " + t.P1.Y.ToString() + " " + t.P1.Z.ToString() + "\n");
                    fs.Write(info, 0, info.Length);
                    info = new System.Text.UTF8Encoding(true).GetBytes("v " + t.P2.X.ToString() + " " + t.P2.Y.ToString() + " " + t.P2.Z.ToString() + "\n");
                    fs.Write(info, 0, info.Length);
                }

                // write texture coords
                //info = new System.Text.UTF8Encoding(true).GetBytes("\n");
                //fs.Write(info, 0, info.Length);
                //info = new System.Text.UTF8Encoding(true).GetBytes("# " + (triangles.Count * 3).ToString() + " Texture Coordinates\n");
                //fs.Write(info, 0, info.Length);
                foreach (Tools.ITexTriangle t in triangles)
                {
                    info = new System.Text.UTF8Encoding(true).GetBytes("vt " + t.T0.X.ToString() + " " + t.T0.Y.ToString() + "\n");
                    fs.Write(info, 0, info.Length);
                    info = new System.Text.UTF8Encoding(true).GetBytes("vt " + t.T1.X.ToString() + " " + t.T1.Y.ToString() + "\n");
                    fs.Write(info, 0, info.Length);
                    info = new System.Text.UTF8Encoding(true).GetBytes("vt " + t.T2.X.ToString() + " " + t.T2.Y.ToString() + "\n");
                    fs.Write(info, 0, info.Length);
                }

                // write normals
                //info = new System.Text.UTF8Encoding(true).GetBytes("\n");
                //fs.Write(info, 0, info.Length);
                //info = new System.Text.UTF8Encoding(true).GetBytes("# " + (triangles.Count * 1).ToString() + " Normals\n");
                //fs.Write(info, 0, info.Length);
                foreach (Tools.ITexTriangle t in triangles)
                {
                    info = new System.Text.UTF8Encoding(true).GetBytes("vn " + t.N.X.ToString() + " " + t.N.Y.ToString() + " " + t.N.Z.ToString() + "\n");
                    fs.Write(info, 0, info.Length);
                }

                // write faces
                //info = new System.Text.UTF8Encoding(true).GetBytes("\n");
                //fs.Write(info, 0, info.Length);
                //info = new System.Text.UTF8Encoding(true).GetBytes("# " + (triangles.Count * 1).ToString() + " Faces\n");
                //fs.Write(info, 0, info.Length);
                info = new System.Text.UTF8Encoding(true).GetBytes("s off\n");
                fs.Write(info, 0, info.Length);
                info = new System.Text.UTF8Encoding(true).GetBytes("g pHMapPlane\n");
                fs.Write(info, 0, info.Length);
                for (int i = 0; i < triangles.Count; i++) {
                    info = new System.Text.UTF8Encoding(true).GetBytes("f " + (i * 3 + 1).ToString() + "/" + (i * 3 + 1).ToString() + "/" + (i + 1).ToString());
                    fs.Write(info, 0, info.Length);
                    info = new System.Text.UTF8Encoding(true).GetBytes("  " + (i * 3 + 2).ToString() + "/" + (i * 3 + 2).ToString() + "/" + (i + 1).ToString());
                    fs.Write(info, 0, info.Length);
                    info = new System.Text.UTF8Encoding(true).GetBytes("  " + (i * 3 + 3).ToString() + "/" + (i * 3 + 3).ToString() + "/" + (i + 1).ToString() + "\n");
                    fs.Write(info, 0, info.Length);
                }

            }
        
        }




        #region String Creator Functions

        private String getParamLine(string paramName, int value)
        {
            return paramName + mSeparator + value.ToString() + "\n";
        }

        private String getParamLine(string paramName, float value)
        {
            return paramName + mSeparator + value.ToString() + "\n";
        }

        private String getParamLine(string paramName, Vector2 value) {
            return paramName + mSeparator + value.X.ToString() + mSeparator + value.Y.ToString() + "\n";
        }


        private String getParamLine(string paramName, Vector3 value) {
            return paramName + mSeparator + value.X.ToString() + mSeparator + value.Y.ToString() + mSeparator + value.Z.ToString() + "\n";
        }
        

        private String getParamLine(string paramName, String value) {
            return paramName + mSeparator + value + "\n";
        }


        private string getParamLineNStore(string paramName, string filePath, Texture2D texture, ImageFileFormat format)
        {
            // save texture
            texture.Save("Content\\" + filePath + "." + format.ToString(), format);

            // generate string line
            //return paramName + mSeparator + "Content\\" + filePath + "." + format.ToString() + "\n";
            return paramName + mSeparator + "Content\\" + filePath + mSeparator + format.ToString() + "\n";
        }



        #endregion



        #region String Parser Functions


        private int parseParameterI(string paramName, string inputLine)
        {
            // create delimiter char array
            char[] delimiterChars = mSeparator.ToCharArray();

            // buffer for parsed words
            string[] words;

            // parse string
            words = inputLine.Split(delimiterChars);



            // convert value
            int value = Convert.ToInt32(words[1]);

            // return value
            return value;
        }


        private float parseParameterF(string paramName, string inputLine)
        {
            // create delimiter char array
            char[] delimiterChars = mSeparator.ToCharArray();

            // buffer for parsed words
            string[] words;

            // parse string
            words = inputLine.Split(delimiterChars);


            // convert value
            float value = (float)Convert.ToDouble(words[1]);

            // return value
            return value;
        }





        private Vector2 parseParameterV2(string paramName, string inputLine)
        {
            // create delimiter char array
            char[] delimiterChars = mSeparator.ToCharArray();

            // buffer for parsed words
            string[] words;

            // parse string
            words = inputLine.Split(delimiterChars);

            // convert value
            Vector2 value = new Vector2((float)Convert.ToDouble(words[1]), (float)Convert.ToDouble(words[2]));

            // return value
            return value;
        }



        private Vector3 parseParameterV3(string paramName, string inputLine)
        {
            // create delimiter char array
            char[] delimiterChars = mSeparator.ToCharArray();

            // buffer for parsed words
            string[] words;

            // parse string
            words = inputLine.Split(delimiterChars);

            // convert value
            Vector3 value = new Vector3((float)Convert.ToDouble(words[1]), (float)Convert.ToDouble(words[2]), (float)Convert.ToDouble(words[3]));

            // return value
            return value;
        }



        private string parseParameterS(string paramName, string inputLine)
        {
            // create delimiter char array
            char[] delimiterChars = mSeparator.ToCharArray();

            // buffer for parsed words
            string[] words;

            // parse string
            words = inputLine.Split(delimiterChars);

            // convert value
            string value = words[1];

            // return value
            return value;
        }



        private Texture2D parseParameterT(string paramName, string inputLine, ContentManager contentManager, GraphicsDevice graphicsDevice)
        {
            // create delimiter char array
            char[] delimiterChars = mSeparator.ToCharArray();

            // buffer for parsed words
            string[] words;

            // parse string
            words = inputLine.Split(delimiterChars);

            // load texture
            Texture2D newText = Texture2D.FromFile(graphicsDevice, words[1] + "." + words[2]); 
                //contentManager.Load<Texture2D>(words[1]);

            // return textue
            return newText;
        }


        #endregion



        

        
    }
}
