// ========================================================================================== //
//             ___  _   _    __    __  __    __    ____  __    ____  _____  _  _              //
//            / __)( )_( )  /__\  (  \/  )  /__\  ( ___)(  )  ( ___)(  _  )( \( )             //
//           ( (__  ) _ (  /(__)\  )    (  /(__)\  )__)  )(__  )__)  )(_)(  )  (              //
//            \___)(_) (_)(__)(__)(_/\/\_)(__)(__)(____)(____)(____)(_____)(_)\_)             //
//            -------------------------------------------- G A M E   E N G I N E              //
//                                                                                            //
// ========================================================================================== //
// Developed and implemented by:                                                              //
//     Mader Theodor                                                                          //
//     Oskam Thomas                                                                           //
//                                                                                            //
// Copyright 2007-2009. All rights reserved.                                                  //
// ========================================================================================== //


using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace HFShader.Camera
{


    // this class implements the camera interface to the renderer and other structures
    // that need access to an implementation of a camera
    public class InterfaceCamera
    {




        ///////////
        // props //
        ///////////


        /// <summary>
        /// Cameras configuration 
        /// </summary>
        protected CamConfiguration mConfiguration = new CamConfiguration();
        public virtual CamConfiguration configuration
        {
            get { return mConfiguration; }
        }


        /// <summary>
        /// viewmatrix
        /// </summary>
        protected Matrix mViewMatrix;
        public virtual Matrix viewMatrix
        {
            get { return mViewMatrix; }
        }


        /// <summary>
        /// projection matrix
        /// </summary>
        protected Matrix mProjectionMatrix;
        public virtual Matrix projectionMatrix
        {
            get { return mProjectionMatrix; }
        }



        /// <summary>
        /// aspect ratio
        /// </summary>
        protected float mAspectRatio = 1280.0f / 720.0f;
        public virtual float aspectRatio
        {
            get { return mAspectRatio; }
            set
            {
                mAspectRatio = value;
                computeProjectionMatrix();
            }
        }



        protected Vector2 mNearFarValues = new Vector2(0.5f, 2100.0f);
        public virtual Vector2 nearFarValues
        {
            get { return mNearFarValues; }
            set
            {
                mNearFarValues = value;
                computeProjectionMatrix();
            }
        }


        /// <summary>
        /// the opening angle of the viewfrustum
        /// </summary>
        protected float mOpeningAngle = 45.0f;
        public virtual float openingAngle
        {
            get { return mOpeningAngle; }
            set
            {
                mOpeningAngle = value;
                computeProjectionMatrix();
                //mProjectionMatrix = Matrix.CreateOrthographic(50f * mAspectRatio, 50f, mNearFarValues.X, mNearFarValues.Y);
            }
        }


        /// <summary>
        /// direction of the camera in xz plane
        /// </summary>
        public virtual Vector2 lookDirXZ
        {
            get
            {
                Matrix rot = Matrix.CreateRotationY(MathHelper.ToRadians(mConfiguration.pitch));
                Vector3 dir = Vector3.Transform(new Vector3(1.0f, 0.0f, 0.0f), rot);
                return new Vector2(dir.X, dir.Z);
            }
        }




        /// <summary>
        /// constructore
        /// </summary>
        public InterfaceCamera() {

            mAspectRatio = 800 / 600;

            mViewMatrix = Matrix.Identity;

            computeProjectionMatrix();

            
            //mProjectionMatrix = Matrix.CreateOrthographic(50f, 24f, mNearFarValues.X, mNearFarValues.Y);

        }



        public virtual void update() {
            #region create Modelview Matrix
            // create modelview matrix
            Matrix rollMatrix = Matrix.CreateRotationX(MathHelper.ToRadians(mConfiguration.roll));
            Matrix pitchMatrix = Matrix.CreateRotationY(MathHelper.ToRadians(mConfiguration.pitch));
            Matrix yawMatrix = Matrix.CreateRotationZ(MathHelper.ToRadians(mConfiguration.yaw));

            Vector3 up = Vector3.Transform(Vector3.UnitY, rollMatrix * yawMatrix * pitchMatrix);
            Matrix transformation = rollMatrix * yawMatrix * pitchMatrix;

            Vector3 lookDir = new Vector3(1.0f, 0.0f, 0.0f);
            lookDir = Vector3.Transform(lookDir, transformation);
            lookDir.Normalize();
            mViewMatrix = Matrix.CreateLookAt(mConfiguration.position, mConfiguration.position + lookDir, up);
            #endregion
        }

        private void computeProjectionMatrix() {
            mProjectionMatrix = Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(mOpeningAngle),
                                                                      mAspectRatio, mNearFarValues.X, mNearFarValues.Y);
            //mProjectionMatrix = Matrix.CreateOrthographic(30 * mAspectRatio, 30, 0.001f, mNearFarValues.Y);
        }

        
    }
}
